@@ -36,7 +36,8 @@ from textwrap import fill
 from traceback import format_exc
 from pyftdi import FtdiLogger
 from pyftdi.eeprom import FtdiEeprom
-from pyftdi.misc import hexdump
+from pyftdi.ftdi import Ftdi
+from pyftdi.misc import add_custom_devices, hexdump
 
 #pylint: disable-msg=too-many-locals
 #pylint: disable-msg=too-many-branches
@@ -54,6 +55,11 @@ def main():
                                help='dump EEPROM content as ASCII')
         argparser.add_argument('-X', '--hexblock', type=int,
                                help='dump EEPROM as indented hexa blocks')
+        argparser.add_argument('-i', '--input', type=FileType('rt'),
+                               help='input ini file to load EEPROM content')
+        argparser.add_argument('-l', '--load', default='all',
+                               choices=('all', 'raw', 'values'),
+                               help='section(s) to load from input file')
         argparser.add_argument('-o', '--output', type=FileType('wt'),
                                help='output ini file to save EEPROM content')
         argparser.add_argument('-s', '--serial-number',
@@ -69,6 +75,9 @@ def main():
                                help='erase the whole EEPROM content')
         argparser.add_argument('-u', '--update', action='store_true',
                                help='perform actual update, use w/ care')
+        argparser.add_argument('-P', '--vidpid', action='append',
+                               help='specify a custom VID:PID device ID, '
+                                    'may be repeated')
         argparser.add_argument('-V', '--virtual', type=FileType('r'),
                                help='use a virtual device, specified as YaML')
         argparser.add_argument('-v', '--verbose', action='count', default=0,
@@ -101,10 +110,17 @@ def main():
             loader = backend.create_loader()()
             loader.load(args.virtual)
 
+        try:
+            add_custom_devices(Ftdi, args.vidpid)
+        except ValueError as exc:
+            argparser.error(str(exc))
+
         eeprom = FtdiEeprom()
         eeprom.open(args.device)
         if args.erase:
             eeprom.erase()
+        if args.input:
+            eeprom.load_config(args.input, args.load)
         if args.serial_number:
             eeprom.set_serial_number(args.serial_number)
         if args.manufacturer:
@@ -113,7 +129,7 @@ def main():
             eeprom.set_product_name(args.product)
         for conf in args.config or []:
             if conf == '?':
-                helpstr = ', '.join(eeprom.properties)
+                helpstr = ', '.join(sorted(eeprom.properties))
                 print(fill(helpstr, initial_indent='  ',
                            subsequent_indent='  '))
                 exit(1)
@@ -141,7 +157,8 @@ def main():
                 hexa = ' '.join(['%02x' % x for x in eeprom.data[pos:pos+16]])
                 print(indent, hexa, sep='')
         if args.update:
-            eeprom.commit(False)
+            if eeprom.commit(False):
+                eeprom.reset_device()
         if args.verbose > 0:
             eeprom.dump_config()
         if args.output:
