@@ -25,16 +25,25 @@
 
 """EEPROM management for PyFdti"""
 
+#pylint: disable-msg=too-many-branches
+#pylint: disable-msg=too-many-locals
+#pylint: disable-msg=wrong-import-position
+#pylint: disable-msg=import-error
+
 import sys
-from binascii import hexlify
+from binascii import hexlify, unhexlify
 from collections import OrderedDict, namedtuple
 from configparser import ConfigParser
-from enum import IntEnum, IntFlag
+from enum import IntEnum
+if sys.version_info[:2] > (3, 5):
+    from enum import IntFlag
 from logging import getLogger
 from random import randint
 from re import match
 from struct import calcsize as scalc, pack as spack, unpack as sunpack
 from typing import BinaryIO, List, Optional, Set, TextIO, Union
+if sys.version_info[:2] == (3, 5):
+    from aenum import IntFlag
 from usb.core import Device as UsbDevice
 from .ftdi import Ftdi, FtdiError
 from .misc import to_bool, to_int
@@ -61,13 +70,18 @@ class FtdiEeprom:
     """
 
     _PROPS = namedtuple('PROPS', 'size user dynoff')
-    """Properties for each FTDI device release."""
+    """Properties for each FTDI device release.
+
+       * size is the size in bytes of the EEPROM storage area
+       * user is the size in bytes of the user storage area, if any/supported
+       * dynoff is the offset in EEPROM of the first bytes to store strings
+    """
 
     _PROPERTIES = {
-        0x0200: _PROPS(0, None, 0x94),     # FT232AM
+        0x0200: _PROPS(0, None, 0),        # FT232AM
         0x0400: _PROPS(256, 0x14, 0x94),   # FT232BM
         0x0500: _PROPS(256, 0x16, 0x96),   # FT2232D
-        0x0600: _PROPS(128, None, 0x98),   # FT232R
+        0x0600: _PROPS(128, None, 0x18),   # FT232R
         0x0700: _PROPS(256, 0x1A, 0x9A),   # FT2232H
         0x0800: _PROPS(256, 0x1A, 0x9A),   # FT4232H
         0x0900: _PROPS(256, 0x1A, 0xA0),   # FT232H
@@ -76,35 +90,35 @@ class FtdiEeprom:
     """EEPROM properties."""
 
 
-    CBUS = IntEnum('CBUS',
-                   'TXDEN PWREN RXLED TXLED TXRXLED SLEEP CLK48 CLK24 CLK12 '
-                   'CLK6 IOMODE BB_WR BB_R', start=0)
+    CBUS = IntEnum('CBus',
+                   'TXDEN PWREN TXLED RXLED TXRXLED SLEEP CLK48 CLK24 CLK12 '
+                   'CLK6 GPIO BB_WR BB_RD', start=0)
     """Alternate features for legacy FT232R devices."""
 
-    CBUSH = IntEnum('CBUSH',
+    CBUSH = IntEnum('CBusH',
                     'TRISTATE TXLED RXLED TXRXLED PWREN SLEEP DRIVE0 DRIVE1 '
-                    'IOMODE TXDEN CLK30 CLK15 CLK7_5', start=0)
+                    'GPIO TXDEN CLK30 CLK15 CLK7_5', start=0)
     """Alternate features for FT232H/FT2232H/FT4232H devices."""
 
-    CBUSX = IntEnum('CBUSX',
+    CBUSX = IntEnum('CBusX',
                     'TRISTATE TXLED RXLED TXRXLED PWREN SLEEP DRIVE0 DRIVE1 '
-                    'IOMODE TXDEN CLK24 CLK12 CLK6 BAT_DETECT BAT_DETECT_NEG '
-                    'I2C_TXE I2C_RXF VBUS_SENSE BB_WR BB_RD TIME_STAMP AWAKE',
+                    'GPIO TXDEN CLK24 CLK12 CLK6 BAT_DETECT BAT_NDETECT '
+                    'I2C_TXE I2C_RXF VBUS_SENSE BB_WR BB_RD TIMESTAMP AWAKE',
                     start=0)
     """Alternate features for FT230X devices."""
 
-    INVERT = IntFlag('INVERT', 'TXD RXD RTS CTS DTR DSR DCD RI')
-    """Inversion flags for FT232R devices."""
+    UART_BITS = IntFlag('UartBits', 'TXD RXD RTS CTS DTR DSR DCD RI')
+    """Inversion flags for FT232R and FT-X devices."""
 
-    CHANNEL = IntFlag('CHANNEL', 'FIFO OPTO CPU FT128 RS485')
+    CHANNEL = IntFlag('Channel', 'FIFO OPTO CPU FT128 RS485')
     """Alternate port mode."""
 
-    DRIVE = IntEnum('DRIVE',
+    DRIVE = IntFlag('Drive',
                     'LOW HIGH SLOW_SLEW SCHMITT _10 _20 _40 PWRSAVE_DIS')
     """Driver options for I/O pins."""
 
-    CFG1 = IntFlag('CFG1', 'CLK_IDLE_STATE DATA_LSB FLOW_CONTROL _08 '
-                            'HIGH_CURRENTDRIVE _20 _40 SUSPEND_DBUS7')
+    CFG1 = IntFlag('Cfg1', 'CLK_IDLE_STATE DATA_LSB FLOW_CONTROL _08 '
+                           'HIGH_CURRENTDRIVE _20 _40 SUSPEND_DBUS7')
     """Configuration bits stored @ 0x01."""
 
     VAR_STRINGS = ('manufacturer', 'product', 'serial')
@@ -118,6 +132,7 @@ class FtdiEeprom:
         self._valid = False
         self._config = OrderedDict()
         self._dirty = set()
+        self._modified = False
 
     def __getattr__(self, name):
         if name in self._config:
@@ -138,7 +153,7 @@ class FtdiEeprom:
         else:
             self._ftdi.open_from_device(device)
         if not ignore:
-            self._read_eeprom()
+            self._eeprom = self._read_eeprom()
             if self._valid:
                 self._decode_eeprom()
 
@@ -164,9 +179,10 @@ class FtdiEeprom:
         self._config = OrderedDict()
         self._dirty = set()
         if not ignore:
-            self._read_eeprom()
+            self._eeprom = self._read_eeprom()
             if self._valid:
                 self._decode_eeprom()
+            self._decode_eeprom()
 
     @property
     def device_version(self) -> int:
@@ -234,7 +250,7 @@ class FtdiEeprom:
            :return: list of CBUS pins
         """
         pins = [pin for pin in range(0, 10)
-                if self._config.get('cbus_func_%d' % pin, '') == 'IOMODE']
+                if self._config.get('cbus_func_%d' % pin, '') == 'GPIO']
         return pins
 
     @property
@@ -251,7 +267,7 @@ class FtdiEeprom:
             cbus = list(range(4))
         mask = 0
         for bix, pin in enumerate(cbus):
-            if self._config.get('cbus_func_%d' % pin, '') == 'IOMODE':
+            if self._config.get('cbus_func_%d' % pin, '') == 'GPIO':
                 mask |= 1 << bix
         return mask
 
@@ -273,6 +289,74 @@ class FtdiEeprom:
             cfg.set('raw', '@%02x' % i, hexa)
         cfg.write(file)
 
+    def load_config(self, file: TextIO, section: Optional[str] = None) -> None:
+        """Load the EEPROM content from an INI stream.
+
+           The ``section`` argument selects which section(s) to load:
+
+           * ``raw`` only loads the raw data (hexabytes) from a previous dump
+           * ``values`` only loads the values section, that is the human
+             readable configuration.
+           * ``all``, which is the default section selection, load the raw
+             section, then overwrite part of it with any configuration value
+             from the ``values`` section. This provides a handy way to use an
+             existing dump from a valid EEPROM content, while customizing some
+             parameters, such as the serial number.
+
+           :param file: input stream
+           :paran section: which section to load from the ini file
+        """
+        self._sync_eeprom()
+        cfg = ConfigParser()
+        cfg.read_file(file)
+        loaded = False
+        sect = 'raw'
+        if section in (None, 'all', sect, ''):
+            if not cfg.has_section(sect):
+                raise FtdiEepromError("No '%s' section in INI file" % sect)
+            options = cfg.options(sect)
+            try:
+                for opt in options:
+                    if not opt.startswith('@'):
+                        raise ValueError()
+                    address = int(opt[1:], 16)
+                    hexval = cfg.get(sect, opt).strip()
+                    buf = unhexlify(hexval)
+                    self._eeprom[address:address+len(buf)] = buf
+            except IndexError:
+                raise ValueError("Invalid address in '%s'' section" % sect)
+            except ValueError:
+                raise ValueError("Invalid line in '%s'' section" % sect)
+            self._compute_crc(self._eeprom, True)
+            if not self._valid:
+                raise ValueError('Loaded RAW section is invalid (CRC mismatch')
+            loaded = True
+        sect = 'values'
+        vmap = {
+            'manufacturer': 'manufacturer_name',
+            'product': 'product_name',
+            'serial': 'serial_number'
+        }
+        if section in (None, 'all', sect, ''):
+            if not cfg.has_section(sect):
+                raise FtdiEepromError("No '%s' section in INI file" % sect)
+            options = cfg.options(sect)
+            for opt in options:
+                value = cfg.get(sect, opt).strip()
+                if opt in vmap:
+                    func = getattr(self, 'set_%s' % vmap[opt])
+                    func(value)
+                else:
+                    try:
+                        self.set_property(opt, value)
+                    except ValueError:
+                        self.log.warning("Ignoring setting '%s': "
+                                         "not implemented", opt)
+            loaded = True
+        if not loaded:
+            raise ValueError('Invalid section: %s' % section)
+        self._sync_eeprom()
+
     def set_serial_number(self, serial: str) -> None:
         """Define a new serial number."""
         self._validate_string(serial)
@@ -307,6 +391,11 @@ class FtdiEeprom:
             self._set_cbus_func(int(mobj.group(1)), value, out)
             self._dirty.add(name)
             return
+        mobj = match(r'([abcd])bus_(drive|slow_slew|schmitt)', name)
+        if mobj:
+            self._set_bus_control(mobj.group(1), mobj.group(2), value, out)
+            self._dirty.add(name)
+            return
         hwords = {
             'vendor_id': 0x02,
             'product_id': 0x04,
@@ -327,7 +416,6 @@ class FtdiEeprom:
             'out_isochronous': (2, 1),
             'suspend_pull_down': (2, 2),
             'has_serial': (2, 3),
-            'has_usb_version': (2, 4),
         }
         if name in confs:
             val = to_bool(value, permissive=False, allow_int=True)
@@ -342,6 +430,13 @@ class FtdiEeprom:
             val = to_int(value) >> 1
             self._eeprom[0x09] = val
             return
+        if name.startswith('invert_'):
+            if not self.device_version in (0x600, 0x1000):
+                raise ValueError('UART control line inversion not available '
+                                 'with this device')
+            self._set_invert(name[len('invert_'):], value, out)
+            self._dirty.add(name)
+            return
         if name in self.properties:
             raise NotImplementedError("Change to '%s' is not yet supported" %
                                       name)
@@ -349,8 +444,9 @@ class FtdiEeprom:
 
     def erase(self) -> None:
         """Erase the whole EEPROM."""
-        self._eeprom = bytearray([0x00] * self.size)
+        self._eeprom = bytearray([0xFF] * self.size)
         self._config.clear()
+        self._dirty.add('eeprom')
 
     def initialize(self) -> None:
         """Initialize the EEPROM with some default sensible values.
@@ -396,14 +492,27 @@ class FtdiEeprom:
 
            :return: True if some changes have been committed to the EEPROM
         """
-        if not self._dirty:
-            self.log.info('No change to commit')
-            return False
-        self.log.info('Changes to commit: %s', ', '.join(sorted(self._dirty)))
         self._sync_eeprom()
+        if not self._modified:
+            self.log.warning('No change to commit, EEPROM not modified')
+            return False
         self._ftdi.overwrite_eeprom(self._eeprom, dry_run=dry_run)
+        if not dry_run:
+            eeprom = self._read_eeprom()
+            if eeprom != self._eeprom:
+                pos = 0
+                for pos, (old, new) in enumerate(zip(self._eeprom, eeprom)):
+                    if old != new:
+                        break
+                pos &= ~0x1
+                raise FtdiEepromError('Write to EEPROM failed @ 0x%02x' % pos)
+            self._modified = False
         return dry_run
 
+    def reset_device(self):
+        """Execute a USB device reset."""
+        self._ftdi.reset(usb_reset=True)
+
     @classmethod
     def _validate_string(cls, string):
         for invchr in ':/':
@@ -453,6 +562,7 @@ class FtdiEeprom:
 
     def _sync_eeprom(self):
         if not self._dirty:
+            self.log.debug('No change detected for EEPROM content')
             return
         if any([x in self._dirty for x in self.VAR_STRINGS]):
             self._generate_var_strings()
@@ -461,31 +571,39 @@ class FtdiEeprom:
         self._update_crc()
         self._decode_eeprom()
         self._dirty.clear()
+        self._modified = True
+        self.log.debug('EEPROM content regenerated (not yet committed)')
 
-    def _compute_crc(self, check=False):
+    def _compute_crc(self, eeprom: Union[bytes, bytearray], check=False):
         mtp = self._ftdi.device_version == 0x1000
-        crc_pos = 0x100 if mtp else len(self._eeprom)
+        crc_pos = 0x100 if mtp else len(eeprom)
         crc_size = scalc('<H')
         if not check:
             # check mode: add CRC itself, so that result should be zero
             crc_pos -= crc_size
-        crc = self._ftdi.calc_eeprom_checksum(self._eeprom[:crc_pos])
+        crc = self._ftdi.calc_eeprom_checksum(eeprom[:crc_pos])
+        if check:
+            self._valid = not bool(crc)
+            if not self._valid:
+                self.log.debug('CRC is now 0x%04x', crc)
+            else:
+                self.log.debug('CRC OK')
         return crc, crc_pos, crc_size
 
     def _update_crc(self):
-        crc, crc_pos, crc_size = self._compute_crc()
+        crc, crc_pos, crc_size = self._compute_crc(self._eeprom, False)
         self._eeprom[crc_pos:crc_pos+crc_size] = spack('<H', crc)
 
-    def _read_eeprom(self):
+    def _read_eeprom(self) -> bytes:
         buf = self._ftdi.read_eeprom(0, eeprom_size=self.size)
-        self._eeprom = bytearray(buf)
-        crc = self._compute_crc(True)[0]
+        eeprom = bytearray(buf)
+        crc = self._compute_crc(eeprom, True)[0]
         if crc:
             if self.is_empty:
                 self.log.info('No EEPROM or EEPROM erased')
             else:
                 self.log.error('Invalid CRC or EEPROM content')
-        self._valid = not bool(crc)
+        return eeprom
 
     def _decode_eeprom(self):
         cfg = self._config
@@ -497,7 +615,6 @@ class FtdiEeprom:
         cfg['self_powered'] = bool(power_supply & (1 << 6))
         cfg['remote_wakeup'] = bool(power_supply & (1 << 5))
         cfg['power_max'] = power_max << 1
-        cfg['has_usb_version'] = bool(conf & (1 << 4))
         cfg['has_serial'] = bool(conf & (1 << 3))
         cfg['suspend_pull_down'] = bool(conf & (1 << 2))
         cfg['out_isochronous'] = bool(conf & (1 << 1))
@@ -506,11 +623,14 @@ class FtdiEeprom:
         cfg['manufacturer'] = self._decode_string(0x0e)
         cfg['product'] = self._decode_string(0x10)
         cfg['serial'] = self._decode_string(0x12)
+        name = None
         try:
-            name = Ftdi.DEVICE_NAMES[cfg['type']]
-            func = getattr(self, '_decode_%s' % name[2:])
+            name = Ftdi.DEVICE_NAMES[cfg['type']].replace('-', '')
+            if name.startswith('ft'):
+                name = name[2:]
+            func = getattr(self, '_decode_%s' % name)
         except (KeyError, AttributeError):
-            pass
+            self.log.warning('No EEPROM decoder for device %s', name or '?')
         else:
             func()
 
@@ -533,16 +653,25 @@ class FtdiEeprom:
             cbus, count, offset, width = cmap[self.device_version]
         except KeyError:
             raise ValueError('This property is not supported on this device')
+        pin_filter = getattr(self,
+                             '_filter_cbus_func_x%x' % self.device_version,
+                             None)
         if value == '?' and out:
-            print(', '.join(sorted([item.name for item in cbus])), file=out)
+            items = {item.name for item in cbus}
+            if pin_filter:
+                items = {val for val in items if pin_filter(cpin, val)}
+            print(', '.join(sorted(items)) if items else '(none)', file=out)
             return
         if not 0 <= cpin < count:
             raise ValueError("Unsupported CBUS pin '%d'" % cpin)
         try:
-            code = cbus[value.upper()]
+            code = cbus[value.upper()].value
         except KeyError:
             raise ValueError("CBUS pin %d does not have function '%s'" %
                              (cpin, value))
+        if pin_filter and not pin_filter(cpin, value.upper()):
+            raise ValueError("Unsupported CBUS function '%s' for pin '%d'" %
+                             (value, cpin))
         addr = offset + (cpin*width)//8
         if width == 4:
             bitoff = 4 if cpin & 0x1 else 0
@@ -556,24 +685,123 @@ class FtdiEeprom:
         self.log.debug('Cpin %d, addr 0x%02x, value 0x%02x->0x%02x',
                        cpin, addr, old, self._eeprom[addr])
 
-    def _decode_230x(self):
+    @classmethod
+    def _filter_cbus_func_x900(cls, cpin: int, value: str):
+        if cpin == 7:
+            # nothing can be assigned to ACBUS7
+            return False
+        if value in 'TRISTATE TXLED RXLED TXRXLED PWREN SLEEP DRIVE0'.split():
+            # any pin can be assigned these functions
+            return True
+        if cpin in (5, 6, 8, 9):
+            # any function can be assigned to ACBUS5, ACBUS6, ACBUS8, ACBUS9
+            return True
+        if cpin == 0:
+            return value != 'GPIO'
+        return False
+
+    @classmethod
+    def _filter_cbus_func_x600(cls, cpin: int, value: str):
+        if value == 'BB_WR':
+            # this signal is only available on CBUS0, CBUS1
+            return cpin < 2
+        return True
+
+    def _set_bus_control(self, bus: str, control: str,
+                         value: Union[str, int, bool],
+                         out: Optional[TextIO]) -> None:
+        if self.device_version == 0x1000:
+            self._set_bus_control_230x(bus, control, value, out)
+            return
+        # for now, only support FT-X devices
+        raise ValueError('Bus control not implemented for this device')
+
+    def _set_bus_control_230x(self, bus: str, control: str,
+                              value: Union[str, int, bool],
+                              out: Optional[TextIO]) -> None:
+        if bus not in 'cd':
+            raise ValueError('Invalid bus: %s' % bus)
+        try:
+            if control == 'drive':
+                candidates = (4, 8, 12, 16)
+                if value == '?' and out:
+                    print(', '.join([str(v) for v in candidates]), file=out)
+                    return
+                value = int(value)
+                if value not in candidates:
+                    raise ValueError('Invalid drive current: %d mA' % value)
+                value //= 4
+                value -= 1
+            elif control in ('slow_slew', 'schmitt'):
+                if value == '?' and out:
+                    print('off, on', file=out)
+                    return
+                value = int(to_bool(value))
+            else:
+                raise ValueError('Unsupported control: %s' % control)
+        except (ValueError, TypeError):
+            raise ValueError('Invalid %s value: %s' % (control, value))
+        config = self._eeprom[0x0c]
+        if bus == 'd':
+            conf = config & 0x0F
+            config &= 0xF0
+            cshift = 0
+        else:
+            conf = config >> 4
+            config &= 0x0F
+            cshift = 4
+        if control == 'drive':
+            conf &= 0b1100
+            conf |= value
+        elif control == 'slow_slew':
+            conf &= 0b1011
+            conf |= value << 2
+        elif control == 'schmitt':
+            conf &= 0b0111
+            conf |= value << 3
+        else:
+            raise RuntimeError('Internal error')
+        config |= conf << cshift
+        self._eeprom[0x0c] = config
+
+    def _set_invert(self, name, value, out):
+        if value == '?' and out:
+            print('off, on', file=out)
+            return
+        if name.upper() not in self.UART_BITS.__members__:
+            raise ValueError('Unknown property: %s' % name)
+        value = to_bool(value, permissive=False)
+        code = getattr(self.UART_BITS, name.upper())
+        invert = self._eeprom[0x0B]
+        if value:
+            invert |= code
+        else:
+            invert &= ~code
+        self._eeprom[0x0B] = invert
+
+    def _decode_x(self):
+        # FT-X series
         cfg = self._config
         misc, = sunpack('<H', self._eeprom[0x00:0x02])
         cfg['channel_a_driver'] = 'VCP' if misc & (1 << 7) else 'D2XX'
-        for bit in self.INVERT:
+        for bit in self.UART_BITS:
             value = self._eeprom[0x0B]
-            cfg['invert_%s' % self.INVERT(bit).name] = bool(value & bit)
+            cfg['invert_%s' % self.UART_BITS(bit).name] = bool(value & bit)
         max_drive = self.DRIVE.LOW | self.DRIVE.HIGH
         value = self._eeprom[0x0c]
         for grp in range(2):
-            conf = value &0xF
-            cfg['group_%d_drive' % grp] = bool((conf & max_drive) == max_drive)
-            cfg['group_%d_schmitt' % grp] = conf & self.DRIVE.SCHMITT
-            cfg['group_%d_slew' % grp] = conf & self.DRIVE.SLOW_SLEW
+            conf = value & 0xF
+            bus = 'c' if grp else 'd'
+            cfg['%sbus_drive' % bus] = 4 * (1+(conf & max_drive))
+            cfg['%sbus_schmitt' % bus] = bool(conf & self.DRIVE.SCHMITT)
+            cfg['%sbus_slow_slew' % bus] = bool(conf & self.DRIVE.SLOW_SLEW)
             value >>= 4
         for bix in range(4):
             value = self._eeprom[0x1A + bix]
-            cfg['cbus_func_%d' % bix] = self.CBUSX(value).name
+            try:
+                cfg['cbus_func_%d' % bix] = self.CBUSX(value).name
+            except ValueError:
+                pass
         cfg['chip'] = Hex2Int(self._eeprom[0x1E])
 
     def _decode_232h(self):
@@ -596,8 +824,14 @@ class FtdiEeprom:
         for bix in range(5):
             value = self._eeprom[0x18 + bix]
             low, high = value & 0x0F, value >> 4
-            cfg['cbus_func_%d' % ((2*bix)+0)] = self.CBUSH(low).name
-            cfg['cbus_func_%d' % ((2*bix)+1)] = self.CBUSH(high).name
+            try:
+                cfg['cbus_func_%d' % ((2*bix)+0)] = self.CBUSH(low).name
+            except ValueError:
+                pass
+            try:
+                cfg['cbus_func_%d' % ((2*bix)+1)] = self.CBUSH(high).name
+            except ValueError:
+                pass
         cfg['chip'] = Hex2Int(self._eeprom[0x1E])
 
     def _decode_232r(self):
@@ -606,17 +840,23 @@ class FtdiEeprom:
         cfg['channel_a_driver'] = 'VCP' if (~cfg0 & (1 << 3)) else ''
         cfg['high_current'] = bool(~cfg0 & (1 << 2))
         cfg['external_oscillator'] = cfg0 & 0x02
-        for bit in self.INVERT:
+        for bit in self.UART_BITS:
             value = self._eeprom[0x0B]
-            cfg['invert_%s' % self.INVERT(bit).name] = bool(value & bit)
+            cfg['invert_%s' % self.UART_BITS(bit).name] = bool(value & bit)
         bix = 0
         while True:
             value = self._eeprom[0x14 + bix]
             low, high = value & 0x0F, value >> 4
-            cfg['cbus_func_%d' % ((2*bix)+0)] = self.CBUS(low).name
+            try:
+                cfg['cbus_func_%d' % ((2*bix)+0)] = self.CBUS(low).name
+            except ValueError:
+                pass
             if bix == 2:
                 break
-            cfg['cbus_func_%d' % ((2*bix)+1)] = self.CBUS(high).name
+            try:
+                cfg['cbus_func_%d' % ((2*bix)+1)] = self.CBUS(high).name
+            except ValueError:
+                pass
             bix += 1
 
     def _decode_2232h(self):
