@@ -30,9 +30,9 @@ import sys
 from importlib import import_module
 from string import printable as printablechars
 from threading import RLock
-from typing import (List, Mapping, NamedTuple, Optional, Sequence, Set,
-                    TextIO, Tuple)
-from urllib.parse import urlsplit, urlunsplit
+from typing import (Any, Dict, List, NamedTuple, Optional, Sequence, Set,
+                    TextIO, Type, Tuple, Union)
+from urllib.parse import SplitResult, urlsplit, urlunsplit
 from usb.backend import IBackend
 from usb.core import Device as UsbDevice, USBError
 from usb.util import dispose_resources, get_string as usb_get_string
@@ -69,6 +69,14 @@ UsbDeviceDescriptor = NamedTuple('UsbDeviceDescriptor',
      * Prefer bus/address selector over index
 """
 
+UsbDeviceKey = Union[Tuple[int, int, int, int], Tuple[int, int]]
+"""USB device indentifier on the system.
+
+   This is used as USB device identifiers on the host. On proper hosts,
+   this is a (bus, address, vid, pid) 4-uple. On stupid hosts (such as M$Win),
+   it may be degraded to (vid, pid) 2-uple.
+"""
+
 class UsbToolsError(Exception):
     """UsbTools error."""
 
@@ -83,9 +91,9 @@ class UsbTools:
     # limitation in pyusb that prevents from opening several times the same
     # USB device. The following dictionary used bus/address/vendor/product keys
     # to track (device, refcount) pairs
-    Devices = {}
     Lock = RLock()
-    UsbDevices = {}
+    Devices = {}  # (bus, address, vid, pid): (usb.core.Device, refcount)
+    UsbDevices = {}  # (vid, pid): {usb.core.Device}
     UsbApi = None
 
     @classmethod
@@ -120,7 +128,7 @@ class UsbTools:
             cls.Lock.release()
 
     @classmethod
-    def flush_cache(cls):
+    def flush_cache(cls, ):
         """Flush the FTDI device cache.
 
            It is highly recommanded to call this method a FTDI device is
@@ -132,7 +140,7 @@ class UsbTools:
            ``Device may have been disconnected``.
         """
         cls.Lock.acquire()
-        cls.UsbDevices = {}
+        cls.UsbDevices.clear()
         cls.Lock.release()
 
     @classmethod
@@ -240,10 +248,33 @@ class UsbTools:
         finally:
             cls.Lock.release()
 
+    @classmethod
+    def release_all_devices(cls, devclass: Optional[Type] = None) -> int:
+        """Release all open devices.
+
+           :param devclass: optional class to only release devices of one type
+           :return: the count of device that have been released.
+        """
+        cls.Lock.acquire()
+        try:
+            remove_devs = set()
+            for devkey in cls.Devices:
+                if devclass:
+                    dev = cls._get_backend_device(cls.Devices[devkey][0])
+                    if dev is None or not isinstance(dev, devclass):
+                        continue
+                dispose_resources(cls.Devices[devkey][0])
+                remove_devs.add(devkey)
+            for devkey in remove_devs:
+                del cls.Devices[devkey]
+            return len(remove_devs)
+        finally:
+            cls.Lock.release()
+
     @classmethod
     def list_devices(cls, urlstr: str,
-                     vdict: Mapping[str, int],
-                     pdict: Mapping[int, Mapping[str, int]],
+                     vdict: Dict[str, int],
+                     pdict: Dict[int, Dict[str, int]],
                      default_vendor: int) -> \
             List[Tuple[UsbDeviceDescriptor, int]]:
         """List candidates that match the device URL pattern.
@@ -255,7 +286,7 @@ class UsbTools:
            :param vdict: vendor name map of USB vendor ids
            :param pdict: vendor id map of product name map of product ids
            :param default_vendor: default vendor id
-           :return: List of (UsbDeviceDescriptor, interface)
+           :return: list of (UsbDeviceDescriptor, interface)
         """
         urlparts = urlsplit(urlstr)
         if not urlparts.path:
@@ -266,8 +297,8 @@ class UsbTools:
 
     @classmethod
     def parse_url(cls, urlstr: str, scheme: str,
-                  vdict: Mapping[str, int],
-                  pdict: Mapping[int, Mapping[str, int]],
+                  vdict: Dict[str, int],
+                  pdict: Dict[int, Dict[str, int]],
                   default_vendor: int) -> Tuple[UsbDeviceDescriptor, int]:
         """Parse a device specifier URL.
 
@@ -315,7 +346,7 @@ class UsbTools:
             vendor, product = desc[:2]
         except IndexError:
             raise UsbToolsError('No USB device matches URL %s' %
-                                urlstr)
+                                urlstr) from None
         if not vendor:
             cvendors = {candidate[0] for candidate in candidates}
             if len(cvendors) == 1:
@@ -336,9 +367,9 @@ class UsbTools:
         return devdesc, interface
 
     @classmethod
-    def enumerate_candidates(cls, urlparts: Tuple[str, str, str, str, str],
-                             vdict: Mapping[str, int],
-                             pdict: Mapping[int, Mapping[str, int]],
+    def enumerate_candidates(cls, urlparts: SplitResult,
+                             vdict: Dict[str, int],
+                             pdict: Dict[int, Dict[str, int]],
                              default_vendor: int) -> \
             Tuple[List[Tuple[UsbDeviceDescriptor, int]], Optional[int]]:
         """Enumerate USB device URLs that match partial URL and VID/PID
@@ -422,8 +453,8 @@ class UsbTools:
 
     @classmethod
     def show_devices(cls, scheme: str,
-                     vdict: Mapping[str, int],
-                     pdict: Mapping[int, Mapping[str, int]],
+                     vdict: Dict[str, int],
+                     pdict: Dict[int, Dict[str, int]],
                      devdescs: Sequence[Tuple[UsbDeviceDescriptor, int]],
                      out: Optional[TextIO] = None):
         """Show supported devices. When the joker url ``scheme://*/?`` is
@@ -450,8 +481,8 @@ class UsbTools:
 
     @classmethod
     def build_dev_strings(cls, scheme: str,
-                          vdict: Mapping[str, int],
-                          pdict: Mapping[int, Mapping[str, int]],
+                          vdict: Dict[str, int],
+                          pdict: Dict[int, Dict[str, int]],
                           devdescs: Sequence[Tuple[UsbDeviceDescriptor,
                                                    int]]) -> \
             List[Tuple[str, str]]:
@@ -463,7 +494,7 @@ class UsbTools:
            :param devdescs: USB devices and interfaces
            :return: list of (url, descriptors)
         """
-        indices = {}
+        indices = {}  # Dict[Tuple[int, int], int]
         descs = []
         for desc, ifcount in sorted(devdescs):
             ikey = (desc.vid, desc.pid)
@@ -488,7 +519,6 @@ class UsbTools:
                     if productids[prdc] == desc.pid:
                         products.append(prdc)
                 if products:
-                    products.sort(key=len)
                     product = products[0]
             except KeyError:
                 pass
@@ -540,9 +570,13 @@ class UsbTools:
                 cls.UsbApi = 1
             else:
                 cls.UsbApi = 2
-        if cls.UsbApi == 2:
-            return usb_get_string(device, stridx)
-        return usb_get_string(device, 64, stridx)
+        try:
+            if cls.UsbApi == 2:
+                return usb_get_string(device, stridx)
+            return usb_get_string(device, 64, stridx)
+        except UnicodeDecodeError:
+            # do not abort if EEPROM data is somewhat incoherent
+            return ''
 
     @classmethod
     def find_backend(cls) -> IBackend:
@@ -587,7 +621,7 @@ class UsbTools:
             # generated device into a list. To save memory, we only
             # back up the supported devices
             devs = set()
-            vpdict = {}
+            vpdict = {}  # Dict[int, List[int]]
             vpdict.setdefault(vendor, [])
             vpdict[vendor].append(product)
             for dev in backend.enumerate_devices():
@@ -622,9 +656,25 @@ class UsbTools:
             cls.UsbDevices[vidpid] = devs
         return cls.UsbDevices[vidpid]
 
+    @classmethod
+    def _get_backend_device(cls, device: UsbDevice) -> Any:
+        """Return the backend implementation of a device.
+
+           :param device: the UsbDevice (usb.core.Device)
+           :return: the implementation of any
+        """
+        try:
+            #pylint: disable-msg=protected-access
+            # need to access private member _ctx of PyUSB device
+            # (resource manager) until PyUSB #302 is addressed
+            return device._ctx.dev
+            #pylint: disable-msg=protected-access
+        except AttributeError:
+            return None
+
     @classmethod
     def _load_backend(cls) -> IBackend:
-        backend = None
+        backend = None  # Optional[IBackend]
         for candidate in cls.BACKENDS:
             mod = import_module(candidate)
             backend = mod.get_backend()
